/*
 *  Copyright (C) 2000 Andrea Bettarini <abettarini@excitecorp.com>
 *
 *  		       Excite Italia
 *  		       http://www.excite.it/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Example of client using UNIX domain datagram protocol.
 */

#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	"ntopquery.h"

#define	MAXLINE	512

sendMessage(sockfd, pserv_addr, servlen, sendline)
     int		sockfd;
     struct sockaddr	*pserv_addr;	/* ptr to appropriate sockaddr_XX structure */
     int		servlen;	/* actual sizeof(*pserv_addr) */
     char		*sendline;
{
  int	n,i;
  char	recvline[MAXLINE + 1];

  n = strlen(sendline);
  printf("Sending: %s\n",sendline);
  if (sendto(sockfd, sendline, n, 0, pserv_addr, servlen) != n)
    printf("sendMessage: sendto error on socket");
  /*
   * Now read a message from the socket and write it to
   * our standard output.
   */

  printf("Waiting for server reply....\n");
  n = recvfrom(sockfd, recvline, MAXLINE, 0,
	       (struct sockaddr *) 0, (int *) 0);
  if (n < 0)
    printf("sendMessage: recvfrom error");
  recvline[n] = 0;	/* null terminate */
  fputs(recvline, stdout);
}

void help(char* program) {
  printf("Usage: %s [-i <ip address>|-m <mac address>|-d <index>|-H]\n", program);  
}

main(argc, argv)
     int	argc;
     char	*argv[];
{
  int			sockfd, clilen, servlen;
  char			*mktemp();
  struct sockaddr_un	cli_addr, serv_addr;
  char message[MAXLINE];
  int index;
  int c;
     
  opterr = 0;
  message[0] = '\0';

  if (argc == 1) {
    help(argv[0]);
    return 1;
  }
     
  while ((c = getopt (argc, argv, "i:m:d:Hh?")) != -1)
    switch (c)
      {
      case 'i':
	snprintf(message, sizeof(message),"findhostbyip %s",optarg);
	break;
      case 'm':
	snprintf(message, sizeof(message),"findhostbymac %s",optarg);
	break;
      case 'd':
	snprintf(message, sizeof(message),"gethostbyindex %s",optarg);
	break;
      case 'H':
	sprintf(message,"hello");
	break;
      case '?':
      case 'h':
	help(argv[0]);
	return 1;
      default:
	abort ();
      }
  
  if(message[0] == '\0') {
    help(argv[0]);
    return(-1);
  }

  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sun_family = AF_UNIX;
  strcpy(serv_addr.sun_path, NTOP_PATH);
  servlen = sizeof(serv_addr.sun_family) + strlen(serv_addr.sun_path);

  if ( (sockfd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
    printf("client: can't open datagram socket");

  bzero((char *) &cli_addr, sizeof(cli_addr));
  cli_addr.sun_family = AF_UNIX;
  strcpy(cli_addr.sun_path, CLI_TMP);
  mktemp(cli_addr.sun_path);
  clilen = sizeof(cli_addr.sun_family) + strlen(cli_addr.sun_path);

  if (bind(sockfd, (struct sockaddr *) &cli_addr, clilen) < 0)
    printf("client: can't bind local address");
	
  sendMessage(sockfd, (struct sockaddr *) &serv_addr, servlen, message);
  close(sockfd);
  unlink(cli_addr.sun_path);
  exit(0);
}

